'''
Outdated file from an outdated project architecture
But ... constant still used 

TODO: move to rstt.config.py
+ doc
+ suited refactoring in ranking.standard and ranking.rating

'''


# elo
DEFAULT_ELO = 1500.0

# glicko
GLICKO_MEAN = 1500.0
GLICKO_VAR = 250.0

# glicko2
GLICKO2_MEAN = 1500.0
GLICKO2_VAR = 250.0
GLICKO2_VOLATILITY = 50.0



import copy
import math


from typeguard import typechecked
import warnings


class Glicko:
    @typechecked
    def __init__(self, minRD: float = 30.0,
                 maxRD: float = 350.0,
                 c: float = 63.2,
                 q: float = math.log(10, math.e)/400,
                 lc: int = 400):
        """Glicko Inferer

        The `Glicko <https://en.wikipedia.org/wiki/Glicko_rating_system>`_ rating system is often described as an improvement of :class:`rstt.ranking.inferer.Elo`.
        here, the implementation is based on Dr. Mark E. Glickman `description <https://www.glicko.net/glicko/glicko.pdf>`_.

        .. note::
            The source paper gives more instruction (notion of rating period) than what an Inferer class should do in RSTT.
            Step1, for example is implemented by the :class:`rstt.ranking.standard.BasicGlicko`
            because it is related to the usage of the system, rather than what the Inferer does.

        .. warning::
            There is no type-checker support for 'Glicko ratings'.
            In the documentation we use the typehint 'GlickoRating'.
            Anything with a public mu and sigma attribute fits the bill.


        Parameters
        ----------
        minRD : float, optional
            minimal value of RD, by default 30.0
        maxRD : float, optional
            maximal value of RD, by default 350.0
        c : float, optional
            constant used for 'inactivity decay', by default 63.2
        q : float, optional
            No idea what it represent, feel free to play arround, by default math.log(10, math.e)/400
        lc : int, optional
            Logistic constant similar to the one in :class:`rstt.rnaking.inferer.Elo`, by default 400
        """

        # model constant
        self.__maxRD = maxRD
        self.__minRD = minRD
        self.lc = lc
        self.C = c
        self.Q = q

    def G(self, rd: float) -> float:
        """_summary_

        Implements: page 3, step2, g(RD) formula.

        Parameters
        ----------
        rd : float
            the RD of a rating

        Returns
        -------
        float
            g(RD)
        """
        return 1 / math.sqrt(1 + 3*self.Q*self.Q*(rd*rd)/(math.pi*math.pi))

    def expectedScore(self, rating1, rating2, update: bool = True) -> float:
        """Compute the expected score

        Implements: page 4, E(s|r,rj,RDj) when update=True
        or page 5, E otherwise.

        Parameters
        ----------
        rating1 : GlickoRating
            'main' rating
        rating2 : GlickoRating
            opponents rating
        update : bool, optional
            Wheter to use the formula for update or not, by default True.

        Returns
        -------
        float
            The expected score of the player with rating1 against player with rating2
        """
        RDi = 0 if update else rating1.sigma
        RDj = rating2.sigma
        ri, rj = rating1.mu, rating2.mu
        return 1 / (1 + math.pow(10, -self.G(math.sqrt(RDi*RDi + RDj*RDj)) * (ri-rj)/400))

    def d2(self, rating1, games: list[tuple[any, float]]) -> float:
        """
        Implements: page 4, d^2 formula.

        Parameters
        ----------
        rating1 : GlickoRating
            the main rating
        games : List[Tuple[GlickoRating, float]]
            A list of [opponent_rating, score_of_rating1]

        Returns
        -------
        float
            the d2 value

        Warns
        -----
            Rarely a ZeroDivisionError occurs. In this case, the warning contains all the computational information.
            Execution continues using a very small value instead.
        """
        all_EJ = []
        all_GJ = []
        for rating2, score in games:
            # get needed variables
            Ej = self.expectedScore(rating1, rating2, update=True)
            RDj = rating2.sigma
            Gj = self.G(RDj)

            # store vairables
            all_EJ.append(Ej)
            all_GJ.append(Gj)

        # big sum
        bigSum = 0
        for Gj, Ej, in zip(all_GJ, all_EJ):
            bigSum += Gj*Gj*Ej*(1-Ej)

        '''
        NOTE:
        Try/Expect is not part of the Glicko official algorithm  presentation.
        But I have encountered Unexpected ZeroDivisionError
        
        This is easly fixed by:
        return 1 / min( self.Q*self.Q*bigSum, lower_bound)
        
        However I could note find any specfic details about the choice of the boundary.
        
        Analytically, the term can not be equal to 0.0, it is always >0.
        Nnumercialy, it happens in extreme situation i.e does not arise in standard 'intended' Glicko usage.
        
        The package is for scientifical experimentation,
        It allows extreme case exploration and can not hide arbitrary choices.

        # !!! Do not fix unless it is possible to link a scientifical source justifying the implementation
        '''
        try:
            # d2 formula
            return 1 / (self.Q*self.Q*bigSum)
        except ZeroDivisionError:
            # !!! BUG: ZeroDivisionError observed with extreme rating differences
            # !!! this will now print variable of interest
            # !!! but code will run assuming maximal and mininal expected value possible between 0 and 1

            # HACK: just assume a very low 'bigSum'
            bigSum = 0.00000000001
            correction = 1 / (self.Q*self.Q*bigSum)

            msg = f"Glicko d2 ERROR: {rating1}, {games}\n {bigSum}, {all_EJ}, {all_GJ}\n d2 return value as been adjusted to 1/{bigSum}"
            warnings.warn(msg, RuntimeWarning)
            return correction

    # TODO: how to typecked
    def prePeriod_RD(self, rating: any) -> float:
        """pre update RD value

        Implements: page 3, step1, formula (b).

        Parameters
        ----------
        rating : GlickoRating
            A rating to 'pre-update'

        Returns
        -------
        float
            the new RD value of the rating. 
        """
        new_RD = math.sqrt(rating.sigma*rating.sigma + self.C*self.C)
        # check boundaries on sigma - ??? move max() elsewhere
        return max(min(new_RD, self.__maxRD), self.__minRD)

    def newRating(self, rating1, games: list[tuple[any, float]]):
        """Rating Update method

        Implements: page 3, step2.

        Parameters
        ----------
        rating1 : GlickoRating
            a rating to update.
        games : List[Tuple[GlickoRating, float]]
            A list of results formated under as [opponent_rating, score_of rating1]

        Returns
        -------
        GlickoRating
            the new updated rating
        """

        # compute term 'a'
        d2 = self.d2(rating1, games)
        a = self.Q / ((1/(rating1.sigma*rating1.sigma)) + (1/d2))

        # lcompute term 'b'
        b = 0
        for rating2, score in games:
            b += self.G(rating2.sigma)*(score -
                                        self.expectedScore(rating1, rating2, update=True))

        # create new rating object to avoid 'side effect'
        rating = copy.copy(rating1)
        # post Period R
        rating.mu += a*b
        # post Period RD
        rating.sigma = math.sqrt(1/((1/rating1.sigma**2) + (1/d2)))

        return rating

    # FIXME: Does not support GameByGame observer
    def rate(self, rating, ratings_opponents: list[any], scores: list[float], *args, **kwars):
        """Glicko rate method

        End to end method to compute a new glicko rating based on a collection of results

        Parameters
        ----------
        rating : GlickoRating
            the rating to update
        ratings : List[GlickoRating]
            list of opponent ratings
        scores : List[float]
            list of score achieved by rating1 against the 'ratings' opponents, in the same order

        Returns
        -------
        GlickoRating
            The new rating.
        """

        # formating
        games = [(r, s) for r, s in zip(ratings_opponents, scores)]
        return self.newRating(rating, games)



"""Modues defining SMatch

SMatch are the notion of game/match, competitive confrontation. It is more of a ponctual event with an outcome rather than a game with its own set of rules and mechanics.
SMatch have participants, accessed by the method :func:`rstt.game.match.Match.players` and the result - a :class:`rstt.stypes.Score` returned by the method :func:`rstt.game.match.Match.scores`
"""


from .match import Match, Duel



from rstt.stypes import SPlayer, RatingSystem
from rstt.ranking.rating import GlickoRating
from rstt.ranking.ranking import Ranking, get_disamb
from rstt.ranking.datamodel import GaussianModel
from rstt.ranking.inferer import Glicko
from rstt.ranking.observer import BatchGame
from rstt.ranking.observer.gameObserver import new_ratings_groups_to_ratings_dict
from rstt.ranking.observer.utils import *

import math


def get_ratings_for_glicko(prior: RatingSystem, data: dict[str, any]) -> None:
    data[RATING] = prior.get(data[TEAMS][0][0])
    data[RATINGS_OPPONENTS] = [prior.get(opponent)
                               for opponent in data[TEAMS][1]]


class BasicGlicko(Ranking):
    def __init__(self, name: str,
                 mu: float = 1500.0, sigma: float = 350.0,
                 minRD: float = 30.0, maxRD: float = 350.0,
                 c: float = 63.2, q: float = math.log(10, math.e)/400,
                 lc: int = 400,
                 players: list[SPlayer] | None = None):
        """Simple Glicko system

        Implement A glicko rating system as originaly `proposed <https://www.glicko.net/glicko/glicko.pdf>`_.

        .. note::
            As recommanded in the source paper, the update() method starts by adjusting each players rating before
            processing any game data (sort of a rating decay)


        Attributes
        ----------
        datamodel: :class:`rstt.ranking.datamodel.GaussianModel` (:class:`rstt.ranking.rating.GlickoRating as rating)
        backend: :class:`rstt.ranking.inferer.Glicko` as backend
        handler :class:`rstt.ranking.observer.BatchGame` as handler


        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        handler : _type_, optional
            Backend as parameter, by default BatchGame()
            The original recommendation is to update the ranking by grouping matches within rating period.
            Which is what the BatchGame Observer do, (each update call represent one period). To match other glicko, use A GameByGame observer
        mu : float, optional
            Datamodel parameter, the default mu of the rating, by default 1500.0
        sigma : float, optional
           Datamodel parameter, the default sigma of the rating, by default 350.0
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """
        super().__init__(name=name,
                         datamodel=GaussianModel(
                             default=GlickoRating(mu, sigma)),
                         backend=Glicko(minRD, maxRD, c, q, lc),
                         handler=BatchGame(),
                         players=players)
        self.handler.query = get_ratings_for_glicko
        self.handler.output_formater = lambda d, x: new_ratings_groups_to_ratings_dict(d, [
                                                                                       [x]])

    @get_disamb
    def __step1(self):
        # TODO: check which player iterator to use
        for player in self:
            rating = self.datamodel.get(player)
            rating.sigma = self.backend.prePeriod_RD(rating)

    def forward(self, *args, **kwargs):
        self.__step1()
        self.handler.handle_observations(
            infer=self.backend, datamodel=self.datamodel, *args, **kwargs)



""" Solver Module

Solver provide a solve(match: SMatch) method to assign a Score to the match. Typicaly a WIN/LOSE/DRAW in case of 'versus' matches
"""


from typing import List, Optional, Callable
from typeguard import typechecked

from rstt import Duel
from rstt.stypes import Score
import rstt.utils.functions as uf

import rstt.config as cfg

import random


'''

    TODO:
    - Extend match to Many-Versus-Many match
    - Extend match to Free-for-all
    - LEVEL_MIXTURES: define differents ways to mix levels in a teams, sum/avg/median/ and set a parameters to tune it solvers
    - Create const value for standard score (maybe enum types) i.e Score.win := [1,0]| Score.lose := [0,1]/ Score.draw := [0.5, 0.5]
    - Work on Score type
    - Add predict() to solvers (and to the stypes.Solver Protocol ?)

'''


WIN = [1.0, 0.0]
"""Default Score Value indicating a 'win' for the first memeber of a 'versus' Match"""
LOSE = [0.0, 1.0]
"""Default Score Value indicating a 'lose' for the first memeber of a 'versus' Match"""
DRAW = [0.5, 0.5]
"""Default Score Value indicating a 'draw' between the two opponents in a 'versus' Match"""


class BetterWin:
    @typechecked
    def __init__(self, with_draw: bool = False):
        """BetterWin Solver

        Implements a deterministic Score generator.
        BetterWin always assign a Win to the best (highest level) participant of a match.

        .. warning::
            Only Supports Duel at the moment.

        Parameters
        ----------
        with_draw : bool, optional
            Wether a draw should be assigned to the game in case of equals level, by default False.
            When False, there is a 'home advantage policy' meaning that in case of equals levels, the first team of the match wins.
        """
        self.with_draw = with_draw

    @typechecked
    def solve(self, duel: Duel, *args, **kwars) -> None:
        level1, level2 = duel.player1().level(), duel.player2().level()
        if level1 > level2:
            score = WIN
        elif level1 < level2:
            score = LOSE
        elif self.with_draw:
            score = DRAW
        else:
            # 'home advantage policy'
            score = WIN
        duel._Match__set_result(result=score)


class ScoreProb:
    @typechecked
    def __init__(self, scores: List[Score], func: Callable[[Duel], Score]):
        """General Purpose Solver

        A ScoreProb

        Parameters
        ----------
        scores : List[Score]
            A list of possible match outcomes.
        func : Callable[[Duel], Score]
            A function taking as input a Duel and producing Score probabilities
        """
        self.scores = scores
        self.probabilities = func

    # ??? can we simply use generic typing, is the code general enough to work for arbitrary SMatch and not just Duel
    # BUG: incompatible func typing in __init__ and .solve() usage -> write a bunch of test
    # FIXME: Should the doc match __init__ signature or the .solve usage ...

    @typechecked
    def solve(self, duel: Duel, *args, **kwars) -> None:
        score = random.choices(population=self.scores,
                               # !!! THE F* is going on here, func should return a Score
                               weights=self.probabilities(duel),
                               k=1)[0]
        duel._Match__set_result(score)


class WeightedScore(ScoreProb):
    @typechecked
    def __init__(self, scores: List[Score], weights: List[float]):
        """Weighted Score assignement

        With this Solver, A score is randomly chosed form a list of options based on weighted.

        Parameters
        ----------
        scores : List[Score]
            A list of possible match outcomes.
        weights : List[float]
            The corresponding weight associated to each Score.

        Raises
        ------
        ValueError
            An error is raised when the scores and weights length are not equal.
        """
        if len(scores) != len(weights):
            msg = f"length of scores ({len(scores)}) does not match length of weights ({len(weights)})"
            raise ValueError(msg)
        super().__init__(scores=scores, func=lambda x: weights)


class CoinFlip(WeightedScore):
    def __init__(self):
        """Random Solver

        Behave like a coin flip, a win or a lose is randomly generated with no regards to any Match details.     
        """
        super().__init__(scores=[WIN, LOSE], weights=[0.5, 0.5])


class BradleyTerry(ScoreProb):
    def __init__(self):
        """Bradley-Terry model

        Implements the famous pairwise model comparaison `probabilistic model <https://en.wikipedia.org/wiki/Bradley–Terry_model>`_.

        It is a ScoreProb Solver where the probability function that a player A with level a, beats a player B with level b, is defined as
            P(A win against B) := a/(a + b)
        """
        super().__init__(scores=[WIN, LOSE], func=self.__probabilities)

    def __probabilities(self, duel: Duel) -> List[float]:
        level1 = duel.teams()[0][0].level()
        level2 = duel.teams()[1][0].level()
        prob = uf.bradleyterry(level1, level2)
        return [prob, 1-prob]


class LogSolver(ScoreProb):
    @typechecked
    def __init__(self, base: Optional[float] = None, lc: Optional[float] = None):
        """Elo like Solver

        The LogSolver implements a standard reparametrization of the Bradley-Terry model that matches Elo rating system.
        In practice it is a ScoreProb with a probability function illustrated on `wismuth <https://wismuth.com/elo/calculator.html>`_.
        FOr a player A with level a, and a Player B with level b, it is defined by the logistic function:
        P(A wins against B) = 1/(1+base^( (b-a) / lc))

        Parameters
        ----------
        base : Optional[float], optional
            The base in the logistic function, by default 10
        lc : Optional[float], optional
            The constant in the logistic function, by default 400

        .. note::
            Default constant in the RSTT package ensure that the LogSolver probabilities matches the expected Score by :class:`rstt.ranking.inferer.Elo`.
            Which means that perfectly accurate predictions are possible when combining both in simulation.
        """
        super().__init__(scores=[WIN, LOSE], func=self.__probabilities)
        self.base = base if base is not None else cfg.LOGSOLVER_BASE
        self.lc = lc if lc is not None else cfg.LOGSOLVER_LC

    def __probabilities(self, duel: Duel) -> List[float]:
        level1 = duel.teams()[0][0].level()
        level2 = duel.teams()[1][0].level()
        prob = uf.logistic_elo(
            base=self.base, diff=level1-level2, constant=self.lc)
        return [prob, 1-prob]



"""Configuration module for rstt simulation default behaviours.

This module contains 'global' variables used as default values in functions and classes across the rstt package.
It offers a simple approach to tune behaviour of simulations once instead of passing identical parameters multiple times in different function calls.

.. warning::
    The module is a work in progress with limited test. Bugs are expected.
    
    
Example
-------
.. code-block:: python
    :linenos:
    
    from rstt import Player
    import rstt.config as cfg
    
    
    cfg.PLAYER_DIST_ARGS['mu'] = 2000
    cfg.PLAYER_DIST_ARGS['sigma'] = 50

    # create players with an average level of approximatively 2000 and a standard deviation of 50.
    players = Player.create(nb=10)
"""

import random


# -------------------- #
# --- Player cfg ----- #
# -------------------- #

# --- BasicPlayer --- #
PLAYER_GAUSSIAN_MU = 1500
"""Deafault mu of :class:`PLAYER_DIST_ARGS`"""

PLAYER_GAUSSIAN_SIGMA = 500
"""Deafault sigma of :class:`PLAYER_DIST_ARGS`"""

PLAYER_DIST = random.gauss
"""Default level generator used by :func:`rstt.player.basicplayer.BasicPlayer.create` when param 'level_dist' is None"""

PLAYER_DIST_ARGS = {'mu': PLAYER_GAUSSIAN_MU,
                    'sigma': PLAYER_GAUSSIAN_SIGMA}
"""Default args for level generator used by :func:`rstt.player.basicplayer.BasicPlayer.create` when param 'level_params' is None"""


# -------------------- #
# --- Player TVS ----- #
# -------------------- #
TIME_SCALE = 100

# --- GaussianPlayer --- #
GAUSSIAN_PLAYER_MEAN_MEAN = 1500
"""Deafault mu of :class:`GAUSSIAN_PLAYER_MEAN_ARGS`"""

GAUSSIAN_PLAYER_MEAN_SIGMA = 500
"""Deafault sigmaof :class:`GAUSSIAN_PLAYER_MEAN_ARGS`"""

GAUSSIAN_PLAYER_SIGMA_MEAN = 100
"""Deafault mu of :class:`GAUSSIAN_PLAYER_SIGMA_ARGS`"""

GAUSSIAN_PLAYER_SIGMA_SIGMA = 50
"""Deafault sigma of :class:`GAUSSIAN_PLAYER_SIGMA_ARGS`"""

GAUSSIAN_PLAYER_MEAN_DIST = random.gauss
"""Default mean level generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'mu' is None"""

GAUSSIAN_PLAYER_SIGMA_DIST = random.gauss
"""Default level standard deviation generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'sigma' is None"""

GAUSSIAN_PLAYER_MEAN_ARGS = {'mu': GAUSSIAN_PLAYER_MEAN_MEAN,
                             'sigma': GAUSSIAN_PLAYER_MEAN_SIGMA}
"""Default args for mean level generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'mu' is None"""

GAUSSIAN_PLAYER_SIGMA_ARGS = {'mu': GAUSSIAN_PLAYER_SIGMA_MEAN,
                              'sigma': GAUSSIAN_PLAYER_SIGMA_SIGMA}
"""Default args for level deviation generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'sigma' is None"""


# --- ExpPlayer --- #
EXPONENTIAL_START_MEAN = 1500
EXPONENTIAL_START_SIGMA = 500
EXPONENTIAL_DIFF_MEAN = 0
EXPONENTIAL_DIFF_SIGMA = 500
EXPONENTIAL_TAU_MEAN = 5*TIME_SCALE
EXPONENTIAL_TAU_SIGMA = 2.5*TIME_SCALE
EXPONENTIAL_START_ARGS = {'mu': EXPONENTIAL_START_MEAN,
                          'sigma': EXPONENTIAL_START_SIGMA}
EXPONENTIAL_DIFF_ARGS = {'mu': EXPONENTIAL_DIFF_MEAN,
                         'sigma': EXPONENTIAL_DIFF_SIGMA}
EXPONENTIAL_TAU_ARGS = {'mu': EXPONENTIAL_TAU_MEAN,
                        'sigma': EXPONENTIAL_TAU_SIGMA}
EXPONENTIAL_START_DIST = random.gauss
EXPONENTIAL_DIFF_DIST = random.gauss
EXPONENTIAL_TAU_DIST = random.gauss


# --- LogisticPlayer --- #
LOGISTIC_START_MEAN = 1500
LOGISTIC_START_SIGMA = 500
LOGISTIC_DIFF_MEAN = 0
LOGISTIC_DIFF_SIGMA = 500
LOGISTIC_CENTER_MEAN = TIME_SCALE / 2
LOGISTIC_CENTER_SIGMA = TIME_SCALE / 4
LOGISTIC_R_MEAN = 0.2
LOGISTIC_R_SIGMA = 0.1
LOGISTIC_START_ARGS = {'mu': LOGISTIC_START_MEAN,
                       'sigma': LOGISTIC_START_SIGMA}
LOGISTIC_DIFF_ARGS = {'mu': LOGISTIC_DIFF_MEAN,
                      'sigma': LOGISTIC_DIFF_SIGMA}
LOGISTIC_CENTER_ARGS = {'mu': LOGISTIC_CENTER_MEAN,
                        'sigma': LOGISTIC_CENTER_SIGMA}
LOGISTIC_R_ARGS = {'mu': LOGISTIC_R_MEAN,
                   'sigma': LOGISTIC_R_SIGMA}
LOGISTIC_START_DIST = random.gauss
LOGISTIC_DIFF_DIST = random.gauss
LOGISTIC_CENTER_DIST = random.gauss
LOGISTIC_R_DIST = random.gauss


# --- CyclePlayer --- #
CYCLE_APPROX_NB_CYCLE = 5
CYCLE_LEVEL_MEAN = 1500
CYCLE_LEVEL_SIGMA = 500
CYCLE_SIGMA_MEAN = 200
CYCLE_SIGMA_SIGMA = 400
CYCLE_TAU_MEAN = TIME_SCALE / CYCLE_APPROX_NB_CYCLE
CYCLE_TAU_SIGMA = TIME_SCALE / (CYCLE_APPROX_NB_CYCLE * 2)
CYCLE_LEVEL_ARGS = {'mu': CYCLE_LEVEL_MEAN,
                    'sigma': CYCLE_LEVEL_SIGMA}
CYCLE_SIGMA_ARGS = {'mu': CYCLE_SIGMA_MEAN,
                    'sigma': CYCLE_SIGMA_SIGMA}
CYCLE_TAU_ARGS = {'mu': CYCLE_TAU_MEAN,
                  'sigma': CYCLE_TAU_SIGMA}
CYCLE_LEVEL_DIST = random.gauss
CYCLE_SIGMA_DIST = random.gauss
CYCLE_TAU_DIST = random.gauss

# --- JumpPlayer --- #
JUMP_LEVEL_MEAN = 1500
JUMP_LEVEL_SIGMA = 500
JUMP_SIGMA_MEAN = 100
JUMP_SIGMA_SIGMA = 50
JUMP_TAU_MEAN = TIME_SCALE / 10
JUMP_TAU_SIGMA = TIME_SCALE / 50
JUMP_LEVEL_ARGS = {'mu': JUMP_LEVEL_MEAN,
                   'sigma': JUMP_LEVEL_SIGMA}
JUMP_SIGMA_ARGS = {'mu': JUMP_SIGMA_MEAN,
                   'sigma': JUMP_SIGMA_SIGMA}
JUMP_TAU_ARGS = {'mu': JUMP_TAU_MEAN,
                 'sigma': JUMP_TAU_SIGMA}
JUMP_LEVEL_DIST = random.gauss
JUMP_SIGMA_DIST = random.gauss
JUMP_TAU_DIST = random.gauss


# -------------------- #
# ---- Match cfg ----- #
# -------------------- #

# tracking game history
MATCH_HISTORY = False
"""Default behaviour of the :class:`rstt.game.match.Match` when param 'tracking' is None.
If set to True, the Match instance will 'try' to add the match to its participants game history.  
"""

DUEL_HISTORY = False
"""Default behaviour of the :class:`rstt.game.match.Duel` when param 'tracking' is None.
If set to True, the Duel instance will 'try' to add the match to its participants game history.  
"""


# -------------------- #
# ---- Solver cfg ---- #
# -------------------- #

LOGSOLVER_BASE = 10
"""Default base of :class:`rstt.solver.solvers.LogSolver` when param 'base' is None"""

LOGSOLVER_LC = 400
"""Default logistic constant of :class:`rstt.solver.solvers.LogSolver` when param 'lc' is None """

# -------------------- #
# --- Competition ---- #
# -------------------- #

# EventStanding Inferer
EVENTSTANDING_DEFAULT_POINTS = {}
"""Default points dictionary of :class:`rstt.ranking.inferer.EventStanding` when param 'default' is None.

The empty dictionary means that if you instanciate an EventStanding without providing a value for 'the default' parameter
and later call the :func:`rstt.ranking.inferer.add_event` without providing a value for the 'points' parameter, you may aswell just not passan 'event' value.
Or even not call the method at all as the corresponding added event will be ignored by the inferer. 
"""



"""Game based Observers
"""

from .obs import ObsTemplate
from .utils import *


class GameByGame(ObsTemplate):
    def __init__(self):
        """Game by Game updating Procedure

        Implementing an iterative approach where each observations triggers the entire updating workflows.
        In particular, new ratings are stored inbetween of each iterations, and the prior ones are lost.

        Observations
        ------------
        game : SMatch, optional
            a game justifying a ranking update, by default None
        games : list[SMatch], optional
            a list of games, by default None
        event : Event, optional
            the observer uses Event.games() to extract the observations, by defualt None
        events: list[Event], optional
            a list of Event, by default None

        Datamodel
        ---------
        Rating: any
            Game based observers make no assumption on ratings type.

        Inferer.rate
        ------------
        teams : list[list[SPlayer]]
        scores : list[float]
        ranks : list[float]
        ratings_groups : list[list[any]]
        """
        super().__init__()
        self.convertor = to_list_of_games
        self.extractor = lambda duels: [duel_data(duel) for duel in duels]
        self.query = get_ratings_groups_of_teams_from_datamodel
        self.output_formater = new_ratings_groups_to_ratings_dict
        self.push = push_new_ratings

    def _set_posteriori(self, *args, **kwargs) -> None:
        # trick: time <=>
        self.posteriori = self.prior


class BatchGame(ObsTemplate):
    def __init__(self):
        """All Matches at once updating procedure

        Alternative to the :class:`rstt.ranking.observer.GamebyGame` observer. Some rating system, like Elo and Glicko
        support updates where all matches are considered at once for the rating update.

        In this workflows, ratings are stored after all matches have been processed. Every computation is performed using the prior ratings
        (i.e the one stored in the datamodel before the method call)

        Observations
        ------------
        game : SMatch, optional
            a game justifying a ranking update, by default None
        games : list[SMatch], optional
            a list of games, by default None
        event : Event, optional
            the observer uses Event.games() to extract the observations, by defualt None
        events: list[Event], optional
            a list of Event, by default None

        Datamodel
        ---------
        Rating: any
            Game based observers make no assumption on ratings type.

        Inferer.rate
        ------------
        teams : list[list[SPlayer]]
        scores : list[float]
        ratings_groups : list[list[any]]
        """
        super().__init__()
        self.convertor = to_list_of_games
        self.extractor = players_records
        self.query = get_ratings_groups_of_teams_from_datamodel
        self.output_formater = new_ratings_groups_to_ratings_dict
        self.push = push_new_ratings



from typing import Optional
from typeguard import typechecked

import rstt.config as cfg
from .playerTVS import PlayerTVS

import random


class GaussianPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: Optional[str] = None, mu: Optional[float] = None, sigma: Optional[float] = None) -> None:
        """Player with a level following a gaussian distribution

        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        mu : float, optional
            The player's mean level,  also considered as the 'original' level. By default None, in this case a level is generated
            using a random distribution :class:`rstt.config.GAUSSIAN_PLAYER_MEAN_DIST with default parameters :class:`rstt.config.GAUSSIAN_PLAYER_MEAN_ARGS`

        sigma : float, optional
            The player's level standard deviation. By default None, in this case a random value is generated
            using a random distribution :class:`rstt.config.GAUSSIAN_PLAYER_SIGMA_DIST with default parameters :class:`rstt.config.GAUSSIAN_PLAYER_SIGMA_ARGS`

        Example
        -------
        The figure below shows a population of 10 players generated with .create() without specifics params.

        .. image:: img/playertvs/ExponentialPlayer.pdf
            :width: 800
        """

        # Genereate if needed a mu value
        mu = mu if mu is not None else cfg.GAUSSIAN_PLAYER_MEAN_DIST(
            **cfg.GAUSSIAN_PLAYER_MEAN_ARGS)

        # pass mu as level to Player
        super().__init__(name=name, level=mu)

        # generate if needed a sigma value
        self.__sigma = sigma if sigma is not None else cfg.GAUSSIAN_PLAYER_SIGMA_DIST(
            **cfg.GAUSSIAN_PLAYER_SIGMA_ARGS)

    def _update_level(self, *args, **kwars) -> None:
        self._PlayerTVS__current_level = random.gauss(
            self._BasicPlayer__level, self.__sigma)



from rstt.stypes import SPlayer, Event
from rstt.ranking import Ranking
from rstt.ranking.datamodel import KeyModel
from rstt.ranking.inferer import EventScoring
from rstt.ranking.observer import PlayerChecker


'''
    TODO: Redesign the ranking concepts
        - ratings as list of achievements
        - KeyModel.ordinal to compute the points (currently EventStanding.rate)
        - backend extracting the relevant achievements of players
        - where goes the  'EventDataSet' component ?
'''


import warnings


class SuccessRanking(Ranking):
    def __init__(self, name: str,
                 window_range: int = 1, tops: int = 1,
                 buffer: int | None = None, nb: int | None = None,
                 players: list[SPlayer] | None = None,
                 default: dict[int, float] | None = None):
        """Merit Based Ranking

        Usefull to implement Ranking system like the one in  `tennis <https://en.wikipedia.org/wiki/ATP_rankings>`_ for example.

        Attributes
        ----------
        datamodel: :class:`rstt.ranking.datamodel.KeyModel` (int as rating)
        backend: :class:`rstt.ranking.inferer.EventScoring`
        handler: :class:`rstt.ranking.observer.PlayerChecker`

        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        buffer : int
            Backend parameter. The number of event to consider for the rating, starting from the last.
        nb : int
            Backend parameter. The actual number of event in the buffer to use for the ratings computation.
        default : Optional[Dict[int, float]], optional
            Backend Parameter. Mapping placement in event to points for the rating, by default None
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """

        if buffer or nb:
            window_range = buffer, tops = nb
            msg = f"buffer and nb will be removed in version 1.0.0, use instead window_range and tops."
            warnings.warn(msg, DeprecationWarning)

        super().__init__(name=name,
                         datamodel=KeyModel(template=int),
                         backend=EventScoring(window_range=window_range,
                                              tops=tops,
                                              default=default),
                         handler=PlayerChecker(),
                         players=players)

    def forward(self, event: Event | None = None, events: list[Event] | None = None):
        new_events = []
        if event:
            new_events.append(event)
        if events:
            new_events += events

        for new_event in new_events:
            self.backend.add_event(new_event)

        self.handler.handle_observations(infer=self.backend,
                                         datamodel=self.datamodel,
                                         players=self.players())



from rstt.ranking import Ranking
from rstt.ranking.datamodel import KeyModel
from rstt.ranking.inferer import PlayerLevel, PlayerWinPRC
from rstt.ranking.observer import PlayerChecker
from rstt.stypes import SPlayer


import numpy as np


class BTRanking(Ranking):
    def __init__(self, name: str = '', players: list[SPlayer] | None = None):
        """Consensus Ranking For the Bradley-Terry Model

        Ranking based on the player's level() method.
        This also work for Time varying player, inherited class from :class:`rstt.player.playerTVS.PlayerTVS`,
        But it needs to be updated manually everytime player's level is updated.


        Attributes
        ----------
        datamodel: :class:`rstt.ranking.datamodel.KeyModel` (float as rating type)
        backend: :class:`rstt.ranking.inferer.PlayerLevel`
        handler: :class:`rstt.ranking.observer.PlayerChecker`


        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        players : _type_, optional
            SPlayer to add to the ranking, by default None

        .. warning::
            BTRanking validity is limited to Bradley-Terry like models and is not suited for simulation using 'None-transitive' level.
        """
        super().__init__(name=name,
                         datamodel=KeyModel(factory=lambda x: x.level()),
                         backend=PlayerLevel(),
                         handler=PlayerChecker(),
                         players=players)


class WinRate(Ranking):
    def __init__(self, name: str,
                 default: float = -1.0,
                 scope: float = np.iinfo(np.int32).max,
                 players: list[SPlayer] | None = None):
        """Ranking based on Win rate


        Ranking that tracks the winrate of :class:`rstt.player.player.Player`.
        The update function does not take any parameters, win rate is computed directly with the player's game history.

        Attributes
        ----------
        datamodel :class:`rstt.ranking.datamodel.KeyModel` (float as rating)
        backend :class:`rstt.ranking.inferer.PlayerWinPRC`
        handler :class:`rstt.ranking.observer.PlayerChecker`


        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        default : float, optional
            A default rating value for when player have no game in their history, by default -1.0
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """
        super().__init__(name,
                         datamodel=KeyModel(default=default),
                         backend=PlayerWinPRC(default=default, scope=scope),
                         handler=PlayerChecker(),
                         players=players)
        # incase player already played games
        self.update()

    def forward(self, *args, **kwargs):
        self.handler.handle_observations(
            datamodel=self.datamodel, infer=self.backend, players=self.players())



from rstt.stypes import SPlayer, SMatch, Event, RatingSystem
from rstt import Duel
# from rstt.ranking.ranking import Ranking
# from rstt.ranking.standing import Standing


from typeguard import typechecked
from typing import Optional

import inspect

# --- Observation's Data --- #
PLAYER = 'player'
PLAYERS = 'players'
TEAM = 'team'
TEAMS = 'teams'
RATINGS_GROUPS = 'ratings_groups'
RATINGS_OPPONENTS = 'ratings_opponents'
RATING = 'rating'
RATING1 = 'rating1'
RATING2 = 'rating2'
RANKS = 'ranks'
SCORES = 'scores'
WEIGTS = 'weights'
NEW_RATINGS = 'new_ratings'


# --- CONVERTOR --- #
@typechecked
def to_list_of_games(game: Optional[SMatch] = None,
                     games: Optional[list[SMatch]] = None,
                     event: Optional[Event] = None,
                     events: Optional[list[Event]] = None):
    observations = []
    if game:
        observations.append(game)
    if games:
        observations += games
    if event:
        observations += event.games()
    if events:
        for ev in events:
            observations += ev.games()
    # NOBUG: user responsability to not pass a given game multiple time (or allow it)
    return observations


@typechecked
def to_list_of_players(player: Optional[SPlayer] = None,
                       players: Optional[list[SPlayer]] = None,
                       team: Optional[SPlayer] = None,
                       teams: Optional[list[SPlayer]] = None,
                       event: Optional[Event] = None,
                       events: Optional[list[Event]] = None
                       # standing: Optional[Standing] = None,
                       # ranking: Optional[Ranking] = None
                       ) -> list[SPlayer]:
    observations = []
    if player:
        observations.append(player)
    if players:
        observations += players
    if team:
        observations.append(team)
    if teams:
        observations += teams
    if event:
        observations += event.participants()
    if events:
        for ev in events:
            observations += ev.participants()
    # if standing:
    #    observations += standing.keys()
    # if ranking:
    #    observations += ranking.players()
    # NOBUG: user responsability to not pass a given player multiple time (or allow it)
    return observations


# --- EXTRACTOR --- #
def duel_data(duel: Duel) -> dict[str, any]:
    # ??? match: Smatch - does it change anything
    # returned value
    data = {}
    # data_points are game summary
    data[TEAMS] = duel.teams()
    data[SCORES] = duel.scores()
    data[RANKS] = duel.ranks()
    return data


def players_records(duels: list[Duel]) -> list[dict[str, any]]:
    # returned value
    datas = []
    for player in active_players(duels):
        # game relevant to the player
        targets = [duel for duel in duels if player in duel]

        # data_points as player 'performance summary'
        data_point = {}
        data_point[TEAMS] = [[player], [
            duel.opponent(player) for duel in targets]]
        data_point[SCORES] = [duel.score(player) for duel in targets]

        datas.append(data_point)
    return datas


# --- QUERY --- #
def get_ratings_groups_of_teams_from_datamodel(prior: RatingSystem, data: dict[str, any]) -> None:
    # inplace data editing
    data[RATINGS_GROUPS] = [[prior.get(player) for player in team]
                            for team in data[TEAMS]]


def get_rating_of_player(prior: RatingSystem, data: dict[str, any]):
    data[RATING] = prior.get(data[PLAYER])


# --- FORMATER --- #
def new_ratings_groups_to_ratings_dict(data: dict[str, any], output: list[list[any]]):
    # inplace data editing
    data[NEW_RATINGS] = {}
    for team, team_ratings in zip(data[TEAMS], output):
        for player, rating in zip(team, team_ratings):
            data[NEW_RATINGS][player] = rating


def new_player_rating(data: dict[str, any], output: any):
    data[NEW_RATINGS] = {data[PLAYER]: output}


# --- PUSH --- #
def push_new_ratings(data: dict[str, any], posteriori: RatingSystem):
    for player, rating in data[NEW_RATINGS].items():
        posteriori.set(player, rating)


# --- Clean Method Calls --- #
def get_function_args(func: callable):
    return inspect.getfullargspec(func).args


def filter_valid_args(args_name: list[str], **kwargs):
    return {key: value for key, value in kwargs.items() if key in args_name}


def call_function_with_args(func: callable, **kwargs):
    func_args = get_function_args(func)
    call_args = filter_valid_args(args_name=func_args, **kwargs)
    return func(**call_args)


# --- Others --- #
def active_players(games: list[SMatch]) -> list[SPlayer]:
    return list(set([player for players in [game.players() for game in games] for player in players]))


# --- no processing --- #
def no_convertion(*args, **kwargs) -> any:
    # !!! probably not what a user expect. BUT what does he expect ?
    return (*args, *kwargs)


def no_extraction(observations: any) -> any:
    return observations


def no_query(prior: RatingSystem, data: any):
    pass


def no_formating(data: any, output: any):
    pass


def no_push(data: any, posteriori: RatingSystem):
    pass



from typing import Dict, Any, Callable, Optional
from typeguard import typechecked
import rstt.config as cfg

import names


class BasicPlayer():
    @typechecked
    def __init__(self, name: Optional[str] = None,
                 level: Optional[float] = None) -> None:
        """Basic Player

        BasicPlayer have a level and a name, that is it. The bare minimum for simulation to run.
        This class is usefull for when your player do not need to track their match history and do not have a time varying level.

        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        level : Optional[float], optional
            The level/skill/strenght of the player. By default None, in this case a level is generated
            using a random distribution :class:`rstt.config.PLAYER_DIST
            with default parameters :class:`rstt.config.PLAYER_DIST_ARGS`
        """
        self.__name = name if name else names.get_full_name()
        self.__level = level if level is not None else cfg.PLAYER_DIST(
            **cfg.PLAYER_DIST_ARGS)

    # --- getter --- #
    def name(self) -> str:
        """Getter method for the name of the player

        Returns
        -------
        str
           The name of the player
        """
        return self.__name

    def level(self) -> float:
        """Getter method for the player's level

        Returns
        -------
        float
            The level of the player.
        """
        return self.__level

    # --- magic methods --- #
    def __repr__(self) -> str:
        return f"Player - name: {self.__name}, level: {self.__level}"

    def __str__(self) -> str:
        return self.__name

    @classmethod
    @typechecked
    def create(cls, nb: int,
               name_gen: Optional[Callable[..., str]] = None,
               name_params: Optional[Dict[str, Any]] = None,
               level_dist: Optional[Callable[..., float]] = None,
               level_params: Optional[Dict] = None):  # -> List[SPlayer]
        """Class method to generate multiple player at once.

        Customizable method to generate a bunch of players with your favorite settings.

        Parameters
        ----------
        nb : int
            The amount of player to create and return.
        name_gen : Callable[..., str], optional
            A name generator. By default None, in this case names are generated using the `names <https://pypi.org/project/names/>`_ package.
        name_params : Dict[str, Any], optional
            Kwargs to pass to the name_gen function, by default None.
        level_dist : Callable[..., float], optional
            A level generator. By default None, in this case it uses a random distribution specifed by :class:`rstt.config.PLAYER_DIST`.
        level_params : Dict, optional
            Kwargs to pass to the level_dist. By default None, in this case it uses default parameters specified by :class:`rstt.config.PLAYER_DIST_ARGS`

        Returns
        -------
        List[BasicPlayer]
            A list of player.
        """
        name_gen = names.get_full_name if name_gen is None else name_gen
        name_params = {} if name_params is None else name_params
        return [cls(name=name_gen(**name_params)) for i in range(nb)]

    @classmethod
    @typechecked
    def seeded_players(cls, nb: int, start: int = 0, inc: float = 100):
        """Create 'seeded' players

        Unlike the :func:`rstt.player.basicplayer.BasicPlayer.create` method, players generated have a deterministic name and level.
        Names are of the form f"Seed_{i}", and the lowest i, the higher the level of the player.

        .. warning::
            The rstt package relies on player's name to identify them - there is no ID. This method can result in name clashing which may lead to confusion and unexpected bahaviour accross simulations.
            It is heavly recommanded to be carefull when calling this method multiple times. Either by tuning the 'start' parameter or deleting previously created player.

        Parameters
        ----------
        nb : int
            The amount of player to create and return.
        start : int, optional
            The first 'i' for the name of players, by default 0.
        inc : float, optional
            The difference of level between playery Seed_i and Seed_i+1, by default 100.

        Returns
        -------
        List[BasicPlayer]
            A list of seeded player in desceding order of level. 
        """
        end = start + (nb * inc)
        levels = list(range(start, end, inc))
        names = [f"Seed_{i}" for i in range(nb, 0, -1)]
        return [cls(name=name, level=level) for name, level in zip(names, levels)]



"""Player based Observers
"""

from .obs import ObsTemplate
from .utils import *


class PlayerChecker(ObsTemplate):
    def __init__(self):
        """Update Procedure based on Player's data

        Iterate over a list of SPlayer, each triggering a inferer.rate()
        call and the new ratings are instantly pushed in the datamodel.
        The assumption is that the player instance itself contains all
        the necessary informations to compute a rating.

        .. caution::
            Prior ratings, stored in the datamodel, are ignored and not passed to the inferer.
            The output of the inferer.rate defines the post ratings,
            which can introduce inconsistency in rating type for the datamodel.

        Observations
        ------------
        player : SMatch, optional
            a game justifying a ranking update, by default None
        players : list[SMatch], optional
            a list of games, by default None
        team : Event, optional
            the observer uses Event.games() to extract the observations, by defualt None
        teams : list[Event], optional
            a list of Event, by default None
        event : Event, optional
            the observer uses Event.games() to extract the observations, by defualt None
        events: list[Event], optional
            a list of Event, by default None

        Datamodel
        ---------
        Rating: any
            Game based observers make no assumption on ratings type.

        Inferer.rate
        ------------
        player: SPlayer
        """
        super().__init__()
        self.convertor = to_list_of_players
        self.extractor = lambda players: [
            {PLAYER: player}for player in players]
        self.query = no_query
        self.output_formater = new_player_rating
        self.push = push_new_ratings



"""Module dedictated to RatingSystem

RatingSystem are ranking component acting as defaultdictionary to store rating for player.
It provide get(), set() and ordinal() methods.
"""

from typing import List, Any, Callable, Optional
from collections import defaultdict
from typeguard import typechecked

from rstt.stypes import SPlayer
from rstt import BasicPlayer

import copy


class keydefaultdict(defaultdict[SPlayer, Any]):
    def __init__(self, default_factory: Callable[[SPlayer], Any]):
        """Defaultdict with default values as function of the missing key.

        This allows rating system to have a default rating based on some data available in a Player instance, such has the level
        or a win rate ratio.

        Parameters
        ----------
        default_factory : Callable[[Player], Any]
            A function computing a default value for a missing key.
        """
        super().__init__()
        self.default_factory = default_factory

    def __missing__(self, key: SPlayer) -> Any:
        # Source: https://stackoverflow.com/a/73975965
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


class KeyModel:
    @typechecked
    def __init__(self, default: Optional[Any] = None, template: Optional[Callable] = None, factory: Optional[Callable] = None, *args, **kwargs):
        r"""Basic Rating system

        The KeyModel is a intuitive implementation of :class:`rstt.stypes.RatingSystem` that strores ratings of player in a defauldict.
        The default rating value can be specified in three different fashions.

        Parameters
        ----------
        default : Any, optional
            A default rating value, by default None.
        template : Callable, optional
            A lambda function to generate default ratings, by default None
        factory : _type_, optional
            A Callable of the form lambda x: ..., where x is the SPlayer with no rating, by default None. If you pass a factory,
            the KeyModel stores ratings in  a :class:`rstt.ranking.datamodel.keydefauldict`.

        .. note::
            If you are using the \'template\' you can additionaly specify \*args and \**kwargs to be passed to the template every time it is called to generate a new rating.
            In the case of the \'factory\', only \**kwargs can be specified. And for the \'default\', no additonal parameters are allowed.

        Raises
        ------
        ValueError
            Error are raised in the case of incompatible parameters. Examples of valid calls are:
            KeyModel(default=1500)
            KeyModel(template=GaussianRating, 1500, 250)
            KeyModel(factory=GaussianRating, mu=1500, sigma=250)
        """
        self.__ratings = self.__init_ratings(
            default, template, factory, *args, **kwargs)
        self.__rtype = self.__get_rating_type()
        self.__default = self.__get_default_rating()

    # --- setter --- #
    @typechecked
    def set(self, key: SPlayer, rating):
        """Set method to manualy assign a rating to a player.

        Parameters
        ----------
        key : SPlayer
            A Player to modify the rating
        rating : Any
            a new rating.

        .. warning::
            The KeyModel class is a 'Generic' container in the sense that it can store any sort of rating, however, all values must be of the same type.
            With the current implementation it is unclear how to enforce it. So it is the responsability of the user to be consistent.

            The set operation will not throw any error. Yet, following calls to :func:`rstt.ranking.datamodel.KeyModel.ordinal` ordinal method could fail without proper error traceback.

            Hopefull future version will fix this issue.
        """
        # TODO: test rating type before assignement and thorw TypeError
        # !!! the problem is how the type is defined? built-in, class, protocol 'attribute protocol' ? isinstance does not work for all cases.
        self.__ratings[key] = rating

    # --- getter --- #
    @typechecked
    def get(self, key: SPlayer) -> Any:
        """getter method for the rating of a player

        Parameters
        ----------
        key : SPlayer
            A player to get the rating

        Returns
        -------
        Any
            The rating of the player
        """
        # QUEST: __getitem__ ?
        return self.__ratings[key]

    def items(self):
        """Dict like items() method

        return the view object returned by the underlying dict.items method

        Returns
        -------
        view object
            items stored in the KeyModel
        """
        return self.__ratings.items()

    def keys(self):
        """Dict like keys() method

        return the view object returned by the underlying dict.keys method

        Returns
        -------
        view object
            keys, i.e. SPlayer.
        """
        return self.__ratings.keys()

    # --- general purpose methords --- #
    def rtype(self) -> type:
        """Getter for the rating type

        .. warning::
            The type of ratings is infered at instanciation of the KeyModel using the type() built-in functions.
            However this can be inconsistant with the user's intention. 

        Returns
        -------
        type
            The rating type.
        """
        return self.__rtype

    def default(self) -> Any:
        """Getter for default rating

        When instanciated, a default rating is generated. In the case of a 'factory' instanciation, the default rating is built with the help of a 'Dummy' :class:`rstt.player.basicplayer.BasicPlayer`

        Returns
        -------
        Any
            The default rating an unrank player gets.
        """
        return self.__default

    def ordinal(self, rating: Any) -> float:
        """Convert a rating into a float value

        The returned value is used to compare players to each other. An higher value is understood as a 'better' player

        .. note::
            This method assume the existence of a magic __float__ method for the ratings.
            If one uses custom ratings object, it is good to either write such method, override the ordinal methods, or design an associated
            :class:`rstt.stypes.RatingSystem`.

        Parameters
        ----------
        rating : Any
            A rating compatible with the KeyModel.rtypes()

        Returns
        -------
        float
            The corresponding value used to compare players in an ordered fashion.
        """

        # REQ: Should not support arg type Player. This is the job of a Ranking/Standing
        # NOTE: name source -> https://fr.wikipedia.org/wiki/Nombre_ordinal
        return float(rating)

    def tiebreaker(self, rating: Any) -> List[Any]:
        """
        .. warning:: 
            DO NOT USE.

            Boilerplate code for future features.

        """
        try:
            return list(rating)
        except:
            return 0

    # --- internal mechanism --- #
    def __init_ratings(self, default, template, factory, *args, **kwargs):
        ''' rating initialization

        return: defaultdict {key: rating}
            where key are the one contained in the standing
            and rating are object used to compute the associated value. 

        REQ:
            - each key as its own rating
            - user can provide a rating object (a) or a type to use as model (b).
            - these to option are imcompatible and an error need to be raised
            - match the default value with the self.__rate_model(rating) method
            (a) We use deepcopy to create a new rating instance with equal value.
            (b) We use a Constructor with params to generate a new rating object.

        TODO:
            - check functool.partial to improve code quaity and readabiltiy.
        '''
        if default != None and not template and not factory:
            if args or kwargs:
                msg = "Can not pass additional argument when using the 'default' parameter."
                raise ValueError(msg)
            ratings = self.__default_ratings(value=default)
        elif template and not default and not factory:
            ratings = self.__template_ratings(template, *args, **kwargs)
        elif factory and not default and not template:
            if args:
                msg = "Can not pass additional positional arguments with the 'factory' parameter specified. Did you mean to use template? Or did you want to pass keyword arguments instead?'"
                raise ValueError(msg)
            ratings = self.__factory_ratings(factory, **kwargs)
        else:
            msg = "Exactly one parameter among 'default', 'template' and 'factory' should be passed."
            raise ValueError(msg)
        return ratings

    def __default_ratings(self, value):
        return defaultdict(lambda: copy.deepcopy(value))

    def __template_ratings(self, template, *args, **kwargs):
        return defaultdict(lambda: template(*args, **kwargs))

    def __factory_ratings(self, func: Callable, **kwargs):
        return keydefaultdict(default_factory=func)

    def __get_rating_type(self):
        dummy = BasicPlayer('dummy', 0.0)
        # !!! This may not always work as intended
        rtype = type(self.__ratings[dummy])
        del self.__ratings[dummy]
        return rtype

    def __get_default_rating(self):
        dummy = BasicPlayer('dummy', 0.0)
        self.__ratings[dummy]
        return self.__ratings.pop(dummy)

    # --- magic methods --- #
    def __delitem__(self, key: SPlayer):
        del self.__ratings[key]

    # ??? __setitem__
    # ??? __getitem__
    # ??? __contain__


class GaussianModel(KeyModel):
    def __init__(self, *arg, **kwargs):
        """Gaussian Rating Systems

        Associated ratings must have a mu and a sigma attributes. 
        """
        super().__init__(*arg, **kwargs)

        '''
        TODO: type check the ratings, something like:
        if not hasattr(self.default(), 'mu') or not hasattr(self.default(), 'sigma'):
            # raise some kind of error
        Or maybe use Protocol with attributes
            
        '''

    # ??? How to typecheck: must at least support trueskill/openskill ratings and rstt GlickoRating
    def ordinal(self, rating) -> float:
        # TODO: cite a good source for the used policy
        """Ordinal method for gaussian values

        The methods implements the mean - 2 * standard deviation policy 

        Parameters
        ----------
        rating : Gaussian like object
            A floating value

        Returns
        -------
        float
            The corresponding value used to compare players in an ordered fashion
        """
        return rating.mu - 2*rating.sigma

    def tiebreaker(self, rating):
        return [rating.mu, rating.sigma]



"""Ranking Module

This module implements utility decorator and a general ranking class.


Glossary
--------

    1. Container Equivalence (Union = Intersection):
    
        - (key in self.datamodel.ratings) <=> (key in self.standing).
        - In the code we refer to 'equivalence'

    2. Rank Disambiguity (point '=' rating):
    
        - self.datamodel.ordinal(key) == self.standing.value(key) for all keys.
        - In the code we refer to 'disambiguity'
    
"""

from typeguard import typechecked
from typing import Any, Union, List, Dict, Callable, Optional

from rstt.ranking import Standing
from rstt.stypes import Inference, RatingSystem, Observer, SPlayer


def set_equi(func: Callable[..., Any]) -> Callable[..., Any]:
    """Equivalence Set Decorator

    Decorator for Ranking methods.
    It enforces the equivalence property after the decorated methods execution

    Parameters
    ----------
    func : Callable[..., Any]
        A method that could alter the equivalence property

    Returns
    -------
    Callable[..., Any]
        A function enforcing the equivalence property
    """

    def wrapper_set(self, *args: Any, **kwargs: Any) -> Any:
        set_action = func(self, *args, **kwargs)
        if self._Ranking__maintain_equivalence:
            self._Ranking__ContainerEquivalence()
        return set_action
    return wrapper_set


def get_equi(func: Callable[..., Any]) -> Callable[..., Any]:
    """Equivalence Get Decorator

    Decorator for Ranking methods.
    It enforces the equivalence property before the decorated methods execution.

    Parameters
    ----------
    func : Callable[..., Any]
        A method that needs 'a-priori' the equivalence property to be satisfied

    Returns
    -------
    Callable[..., Any]
        A function with the expected behaviour.
    """

    def wrapper_get(self, *args: Any, **kwars: Any) -> Any:
        if self._Ranking__maintain_equivalence:
            self._Ranking__ContainerEquivalence()
        return func(self, *args, **kwars)
    return wrapper_get


def get_disamb(func: Callable[..., Any]) -> Callable[..., Any]:
    """Disambiguity Get Decorator

    Decorator for ranking methods.
    It enforces the disambuguity property before the decorated method execution.

    Parameters
    ----------
    func : Callable[..., Any]
        A method that needs 'a-priori' the disambuguity property to be satisfied.

    Returns
    -------
    Callable[..., Any]
        A function with the expected behaviour.
    """

    def wrapper_get(self, *args: Any, **kwars: Any) -> Any:
        if self._Ranking__maintain_disambiguity:
            self._Ranking__RankDisambiguity()
        return func(self, *args, **kwars)
    return wrapper_get


def set_disamb(func: Callable[..., Any]) -> Callable[..., Any]:
    """Disambiguity Set Decorator

    Decorator for ranking methods.
    It enforces the disambuguity property after the decorated method execution.

    Parameters
    ----------
    func : Callable[..., Any]
        A method that could alter the disambuguity property.

    Returns
    -------
    Callable[..., Any]
        A function enforcing the disambiguity property
    """

    def wrapper_set(self, *args: Any, **kwargs: Any) -> Any:
        set_action = func(self, *args, **kwargs)
        if self._Ranking__maintain_disambiguity:
            self._Ranking__RankDisambiguity()
        return set_action
    return wrapper_set


class Ranking():
    @typechecked
    def __init__(self, name: str,
                 datamodel: RatingSystem,
                 backend: Inference,
                 handler: Observer,
                 players: Optional[List[SPlayer]] = None):
        """Ranking for players

        The rstt package implements its own definition of a ranking.

        Formally an rstt ranking consist of a

            - A standing: An ordered sequence of players associated with an indication of their skills.
            - A rating system - storing player's rating.
            - A statistical inference system - set of equations.
            - An update procedure

        And two 'hidden' notions:

            - Observables: the set of 'update triggers' justifying a change of ratings. (what can be processed by the handler)
            - An ordinal function converting rating into float values (provided by the rating system)

        Parameters
        ----------
        name : str
            A name to identify the ranking
        datamodel : RatingSystem
            A container storing rating of players and providing an orinal() funtion to convert rating into floating values.
        backend : Inference
            The 'math' behind the ranking system.
        handler : Observer
            A workflow handling the ranking update procedure
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """

        # name/identifier - usefull for plot
        self.name = name

        # fundamental notions of the Ranking Class
        self.standing = Standing()
        self.backend = backend
        self.datamodel = datamodel
        self.handler = handler

        # state control variable
        self.__equivalence = True
        self.__disambiguity = True

        # protocol control variable
        self.__maintain_equivalence = True
        self.__maintain_disambiguity = True

        if players:
            self.add(keys=players)

    # --- Containers standard methods --- #
    @set_disamb
    @set_equi
    @typechecked
    def add(self, keys: List[SPlayer]):
        """Add players to the ranking

        Each Player receive a default rating by the datamodel.
        It is possible to manipulate it using the set_rating() method.

        Parameters
        ----------
        keys : List[SPlayer]
            Players to be ranked.
        """
        # turn off maintainance for optimization
        should_maintain = self.__maintain_equivalence
        self.__maintain_equivalence = False

        # perform iteratively addition
        for key in keys:
            self.__add(key)

        # restaure Ranking status
        self.__maintain_equivalence = should_maintain

    def __add(self, key: SPlayer):
        if key in self:
            msg = f'Can not add a key already present in the Ranking, {key}'
            raise ValueError(msg)
        # default dict get operator for missing key
        self.datamodel.get(key)

        # self.datamodel do not match self.standing
        self.__equivalence = False

    # --- magic methods --- #
    def __getitem__(self, *args, **kwargs) -> Union[SPlayer, List[SPlayer], int, List[int]]:
        ''' get item based on a rank or a key

        NOBUG:
            - sorting is handled by the Standing itself
            - so is typechecking
        '''

        return self.standing.__getitem__(*args, **kwargs)

    def __delitem__(self, key: SPlayer):
        ''' delete element from the Ranking

        REQ:
            - element needs to be remove both from the standing and the RatingSystem

        NOBUG:
            - del standing[key] is typechecked. CALL MUST BE BEFORE RatingSystem

        # ???:
            - could del succeed on standing but fail on RatingSystem.
            This would potentialy lead to an invalid ranking state because __delitem__ is not decorated.
            A ranking invalid state can exactly be the reason why this scenario happens.
            What is the best approach to this situation ?
        '''
        del self.standing[key]
        del self.datamodel[key]

    def __contains__(self, key: SPlayer):
        '''
        NOTE:
            - it does match standing behavior as specified but
            depending on the choices, 'p in self' can be slower than
            'p in self.standing' / 'p in self.datamodel'
        '''
        return key in self.standing

    def __len__(self):
        return len(self.standing)

    def __iter__(self):
        return self.standing.__iter__()

    # --- getter --- #
    def rank(self, player: SPlayer) -> int:
        """Getter for player Rank

        Equivalent to ranking.standing[player].


        Parameters
        ----------
        player : SPlayer
            A player to get his rank.

        Returns
        -------
        int
            The rank of the player
        """
        return self[player]

    def ranks(self, players: List[SPlayer]) -> List[int]:
        """Getter for players Rank

        Equivalent to ranking.standing[players].


        Parameters
        ----------
        players : SPlayer
            Player to get their ranks.

        Returns
        -------
        List[int]
            The corresponding ranks of the players
        """
        return [self.rank(player) for player in players]

    def rating(self, player: SPlayer) -> Any:
        """Get method for rating

        Rating object is the internal model associated to a key.
        Ratings are used to automaticly compute values for the sorting feature of a Standing.

        Parameters
        ----------
        player : Player
            A key in the Ranking

        Returns
        -------
        Any
            The associated model to the provided key. The type is defined by Ranking.RatingSystem.rtype

        Raises
        ------
        KeyError
        """
        if player in self:  # NOBUG RatingSystem is a defaultdict
            return self.datamodel.get(player)
        else:
            msg = f"{player} is not present in {self.standing}"
            raise KeyError

    def ratings(self) -> List[Any]:
        """Get method for all ratings

        Returns
        -------
        list[Any]
            A list of all rating object present in the Ranking, in order of the Standing.
        """
        return [self.rating(player) for player in self]

    def players(self) -> List[SPlayer]:
        """Get method of all keys

        Alias for Ranking.standing.keys()

        Returns
        -------
        List[Player]
            A list of all player in descending order of their associated values.
        """
        return self.standing.keys()

    def point(self, player: SPlayer) -> float:
        """Get the point associated to a key

        Alias for Ranking.standing.value(player)

        Returns
        -------
        float
            the associated value.
        """
        return self.standing.value(player)

    def points(self) -> List[float]:
        """Get method of all values

        Alias for Ranking.standing.values()

        Returns
        -------
        List[float]
            A list of all associated values in descending order.
        """
        return self.standing.values()

    def status(self) -> Dict[str, Union[bool, str]]:
        """Get ranking's control variables

        This method can be usefull for debugging purposes.
        However, there should be no reason to use it unless a user intentionnaly manipulate the ranking internal mechanism.


        Returns
        -------
        Dict
            name of control variables with their current values.

        :meta private:
        """
        return {'equivalence': self.__equivalence,
                'disambiguity': self.__disambiguity,
                'maintain_equivalence': self.__maintain_equivalence,
                'm_disambanbiguity': self.__maintain_disambiguity}

    # ??? items() -> List[(rank, player, ratings)]
    # ??? item(key) -> (rank, player, ratings)
    # --- setter --- #
    @set_disamb
    @set_equi
    def set_rating(self, key: SPlayer, rating: Any):
        """A method to assign a rating to a Player


        The Ranking delegate this task to a 'RatingSystem' instance stored as attribute 'rankink.datamodel'.
        The RatingSystem define what rating type is accepted and wether a set operation is authorized for the provided key.

        Parameters
        ----------
        key : Player
            A Player
        rating : Any
            A rating object associated to the key
        """
        self.datamodel.set(key, rating)
        self.__equivalence = False

    # ??? remove()
    # ??? pop()

    # --- general purpose methods --- #
    @get_disamb
    @get_equi
    def plot(self):
        """Plot method

        Display the ranking to the standard output
        """
        self.standing.plot(standing_name=self.name)

    @set_disamb
    @set_equi
    def update(self, *args, **kwargs):
        """Update method

        Core functionality of the ranking class allowing player's rating to change and ranks to change.
        It accept aribitrary parameters.

        .. nwarning::
            This method in itself does not do anything. It is a wrapper ensuring any update do not alterate the internal state of the ranking.
            In no cases should this be overriden. Instead refer to :func:`rstt.ranking.ranking.Ranking.forward`
        """
        self.forward(*args, **kwargs)

        # NOTE: How do we know if the ranking state changed ?
        # HACK: always assume it did
        self.__disambiguity = False
        self.__equivalence = False

    def forward(self, *args, **kwargs):
        """Internal 'update' function

        This method calls the handler :func:`rstt.stypes.handle_observations` with the parameters of the update function.

        .. note::
            FOR RANKING DESIGNER ONLY
            method designed for devellopers who wants to modify the ranking.update function's behavivous.
            In most cases, it is sufficient to write an apropriate observer as the ranking.handler.

            However, sometimes it is relevant to do some ranking preprocessing before any rating updates.
            This would not always be possible to do inside the handle_observations method as the observer do not have access to all ranking attributes.
        """
        self.handler.handle_observations(infer=self.backend,
                                         datamodel=self.datamodel,
                                         *args, **kwargs)

    @set_equi
    @set_disamb
    @get_disamb
    @get_equi
    @typechecked
    def rerank(self, permutation: List[int], name: str = None, direct: bool = True):
        """Reorder the ranking.

        Inplace modification of the ranking state by reordering the players while maintaining a coherent state.
        This means that each player will be re assigned a new rating corresponding to the desired permuatation

        Parameters
        ----------
        permutation : List[int]
            A permutation of the ranking indices.
        name : str, optional
            A name, by default None
        direct : bool, optional
            Wether to apply the permutation directly, or its inverse, by default True

        Raises
        ------
        ValueError
            When the permutation is not a permutation over the ranking indices.
        """

        # check permutation validity
        if len(self) != len(permutation):
            # NOTE: without this check, the code will raise an IndexError and could be harder for the user to understand what went wrong.
            msg = f"permutation must be a list of len {len(self)}"
            raise ValueError
        if not (set(permutation) == set(list(range(len(self))))):
            msg = f"permutation must contain each value from 0 to {len(self)-1} exactly once"
            raise ValueError(msg)

        # rename the ranking:
        if name is not None:
            self.name = name

        pairs = []
        for current_rank, future_rank in enumerate(permutation):
            if not direct:
                current_rank, future_rank = future_rank, current_rank
            player = self[current_rank]
            ratings = self.datamodel.get(self[future_rank])
            pairs.append((player, ratings))
        for p, r in pairs:
            self.datamodel.set(p, r)

    @get_disamb
    @get_equi
    def fit(self, players: List[SPlayer]) -> Standing:
        seeding = Standing(default=self.standing._Standing__default,  # ??? lower instead of default
                           lower=self.standing._Standing__min,
                           upper=self.standing._Standing__max,
                           protocol=self.standing._Standing__protocol)  # Optimize
        points = [self.point(
            player) if player in self else None for player in players]
        seeding.add(players, points)
        return seeding

    # --- internal mechanism --- #

    def __ContainerEquivalence(self):
        ''' property checker'''

        # get keys
        standing_keys = set(self.standing.keys())
        RatingSystem_keys = set(self.datamodel.keys())

        if standing_keys == RatingSystem_keys:
            self.__equivalence = True
        elif standing_keys <= RatingSystem_keys:  # a <= b means a.issubset(b)
            # a - b means a.difference(b)
            not_ranked_players = list(RatingSystem_keys - standing_keys)
            new_points = []
            for player in not_ranked_players:
                # NOBUG: player is in the RatingSystem. 'get()' is safe to perform.
                new_points.append(self.datamodel.ordinal(
                    self.datamodel.get(player)))
            # NOBUG: no ambiguity is introduce this way
            self.standing.add(keys=not_ranked_players, values=new_points)
            self.__equivalence = True
        else:
            # TODO: write a good error message
            msg = f''
            raise RuntimeError(msg)

    def __RankDisambiguity(self):
        ''' property checker'''
        for player in self.standing:
            # TODO: performance check (a) assign only if needed (b) assign always
            # ??? (a / b) as user options ?
            rating = self.datamodel.get(player)
            point = self.datamodel.ordinal(rating)
            if self.standing.value(player) != point:
                self.standing[player] = point

        self.__disambiguity = True



from rstt.stypes import Inference, Observer, RatingSystem
from ..datamodel import KeyModel
from . import utils as uo


class ObsTemplate(Observer):
    def __init__(self):
        self.prior: RatingSystem = None
        self.posteriori: RatingSystem = None

    # --- Observer 'Parameters' --- #
    def convertor(self, *args, **kwargs) -> any:
        raise NotImplementedError

    def extractor(self, observations: any) -> list[any]:
        raise NotImplementedError

    def query(self, prior: RatingSystem, data: any):
        raise NotImplementedError

    def output_formater(self, data: any, output: any):
        raise NotImplementedError

    def push(self, data: any, posteriori: RatingSystem):
        raise NotImplementedError

    # --- Observer Main Function --- #
    def handle_observations(self, infer: Inference, datamodel: RatingSystem, *args, **kwargs) -> None:
        # observer initialization
        self._handling_start(datamodel)

        # data transformation
        observations = self.convertor(*args, **kwargs)
        data = self.extractor(observations)
        # process each 'rate-trigger'
        for data_point in data:
            # get corresponding priors
            self.query(self.prior, data_point)

            # perofrm rating evaluation
            self.output_formater(
                data_point, uo.call_function_with_args(infer.rate, **data_point))

            # store posteriori
            self.push(data_point, self.posteriori)

        # terminate the process
        self._handling_end(datamodel)

    # --- internal mechanism --- #
    def _set_prior(self, datamodel: RatingSystem) -> None:
        self.prior = datamodel

    def _set_posteriori(self, datamodel: RatingSystem) -> None:
        self.posteriori = KeyModel(default=0)

    def _handling_start(self, datamodel: RatingSystem):
        self._set_prior(datamodel)
        self._set_posteriori(datamodel)

    def _handling_end(self, datamodel: RatingSystem):
        for player, post_rating in self.posteriori.items():
            datamodel.set(key=player, rating=post_rating)
        self.prior = None
        self.posteriori = None


class NoHandling(ObsTemplate):
    def __init__(self):
        """
        :exclude-members:
        """
        super().__init__()
        self.extractor = uo.no_extraction
        self.convertor = uo.no_convertion
        self.query = uo.no_query
        self.output_formater = uo.no_formating
        self.push = uo.no_push



from rstt.ranking import Ranking
from rstt.ranking.datamodel import KeyModel
from rstt.ranking.inferer import Elo
from rstt.ranking.observer import GameByGame
from rstt.stypes import SPlayer

from typeguard import typechecked


class BasicElo(Ranking):
    def __init__(self, name: str, default: float = 1500,
                 k: float = 20.0,
                 lc: float = 400.0,
                 base: float = 10.0,
                 players: list[SPlayer] | None = None):
        """Simple Elo System


        Attributes
        ----------
        datamodel: :class:`rstt.ranking.datamodel.KeyModel` (float as rating type)
        backend: :class:`rstt.ranking.inferer.Elo`
        handler: :class:`rstt.ranking.observer.GameByGame`


        Parameters
        ----------
        name : str, optional
            A name to identify the ranking, by default ''
        default : float, optional
            Datamodel parameter, a default elo rating, by default 1500.0
        k : float, optional
            Backend parameter, the K value, by default 20.0
        lc : float, optional
            Backend parameter, constant dividing the ratings difference in the expected score formula , by default 400.0
        players : Optional[List[SPlayer]], optional
            Players to register in the ranking, by default None
        """
        super().__init__(name=name,
                         datamodel=KeyModel(default=default),
                         backend=Elo(k=k, lc=lc, base=base),
                         handler=GameByGame(),
                         players=players)



"""Implement classic and usefull Ranking System
"""

from .consensus import BTRanking, WinRate
from .basicOS import BasicOS
from .basicElo import BasicElo
from .basicGlicko import BasicGlicko
from .successRanking import SuccessRanking



"""Modules for ranking purposes
"""

from . import rating

from .standing import Standing
from .ranking import Ranking

from .datamodel import KeyModel, GaussianModel

from .observer import (
    GameByGame, BatchGame,
    PlayerChecker, NoHandling
)
from .inferer import (
    Elo, Glicko,
    PlayerLevel, PlayerWinPRC, EventScoring
)

from .standard import (
    BTRanking, WinRate, SuccessRanking,
    BasicElo, BasicGlicko, BasicOS,
)



from typing import List, Tuple, Any


# ----------------- #
# --- Shuffling --- #
# ----------------- #
def riffle_shuffle(half1: List[Any], half2: List[Any]) -> List[Any]:
    return [half[i] for i in range(len(half1)) for half in [half1, half2]]


# ----------------- #
# --- Splitting --- #
# ----------------- #
def symetric_split(elems: List[Any]) -> Tuple[List[Any], List[Any]]:
    h = len(elems)//2
    half1 = elems[:h]
    half2 = list(reversed(elems[-h:]))
    return half1, half2


def middle_split(elems: List[Any]) -> Tuple[List[Any], List[Any]]:
    h = len(elems)//2
    half1 = elems[:h]
    half2 = elems[-h:]
    return half1, half2


def neighboor_split(elems: List[Any]) -> Tuple[List[Any], List[Any]]:
    half1 = elems[::2]
    half2 = elems[1::2]
    return half1, half2


# ---------------- #
# --- matching --- #
# ---------------- #
def symetric_match(elems: List[Any]) -> List[List[Any]]:
    return [[elems[i], elems[-i]] for i in range(len(elems)//2)]


def parallel_match(elems: List[Any]) -> List[List[Any]]:
    h = len(elems)//2
    return [[elems[i], elems[h+i]] for i in range(h)]


def neighboor_match(elems: List[Any]) -> List[List[Any]]:
    return [[elems[2*i], elems[2*i+1]] for i in range(len(elems)//2)]


# --------------- #
# --- Common ---- #
# --------------- #
def ruban(players: List[Any]) -> List[List[Any]]:
    if len(players) % 2 == 0:
        return _ruban_even(players)
    else:
        return _ruban_odd(players)


def _ruban_even(players: List[Any]) -> List[List[Any]]:
    # QUEST: does it work with odd length of players input ?
    # QUEST: return List[List[List[Any]]] to match other matching func logic ?
    '''
    implement 'clock-like' algorithm, source:
    https://en.wikipedia.org/wiki/File:Round-robin_tournament_10teams_en.png
    '''

    # prevent side-effect due to .pop() usage
    players = [player for player in players]

    # return value
    rounds = []

    # algorithm variable - num10 in the ref .png
    fix = players.pop(len(players)-1)

    # control variables
    r = 1  # current round
    # expected len(rounds) - total amount of rounds
    nb = len(players)-1 if len(players) % 2 == 0 else len(players)

    # build rounds one-by-one
    while r <= nb:

        # round[r-1]
        new_round = []
        # num1, 2, 3, ...
        flex = players.pop(0)

        # first we deal with the special match
        if r % 2 == 0:
            # duel := Duel(flex, fix)
            new_round.append(flex)
            new_round.append(fix)
        else:
            # duel := Duel(fix, flex)
            new_round.append(fix)
            new_round.append(flex)

        # then with the ruban
        half = len(players)//2
        half1 = players[:half]
        half2 = players[half:]
        half2.reverse()
        for index, (p1, p2) in enumerate(zip(half1, half2)):
            if index % 2 == 0:
                # duel = Match(p2,p1)
                new_round.append(p2)
                new_round.append(p1)
            else:
                # duel = Match(p1,p2)
                new_round.append(p1)
                new_round.append(p2)

        # update variable
        rounds.append(new_round)
        r += 1
        players.append(flex)

    return rounds


def _ruban_odd(players: List[Any]) -> List[List[Any]]:
    # each player get a BYE round
    BYE = None
    # avoid side- effect
    players = [player for player in players]
    players.append(BYE)
    rounds = _ruban_even(players)
    for round in rounds:
        for p1, p2 in zip(*neighboor_split(round)):
            if not p1 or not p2:
                round.remove(p1)
                round.remove(p2)

    return rounds


def chord_diagrams_n6(players: List[Any]) -> List[List[Any]]:
    '''
    source:
    https://en.wikipedia.org/wiki/Double_factorial
    https://en.wikipedia.org/wiki/Chord_diagram_(mathematics)

    NOTE: I do not know a algorithm to compute it for an abritrary n=2*k (in reseaonable complexity)
    But n=6 is a frequent needs in SwissBracket for 16 players variations so I hard coded the 15 solutions
    '''

    if len(players) != 6:
        msg = f'chord_diagrams_n6 function is implemented only for 6 participants, len(players) = {len(players)})'
        raise ValueError(msg)

    matchings = [
        [players[0], players[5], players[1], players[4], players[2], players[3]],
        [players[0], players[5], players[1], players[5], players[2], players[4]],
        [players[0], players[4], players[1], players[5], players[2], players[3]],
        [players[0], players[2], players[1], players[4], players[3], players[5]],
        [players[0], players[3], players[1], players[4], players[2], players[5]],

        [players[0], players[1], players[2], players[5], players[3], players[4]],
        [players[0], players[2], players[1], players[2], players[4], players[5]],
        [players[0], players[4], players[1], players[2], players[3], players[5]],
        [players[0], players[3], players[1], players[5], players[2], players[4]],
        [players[0], players[5], players[1], players[2], players[3], players[4]],

        [players[0], players[3], players[1], players[2], players[4], players[5]],
        [players[0], players[2], players[1], players[5], players[3], players[4]],
        [players[0], players[1], players[2], players[4], players[3], players[5]],
        [players[0], players[4], players[1], players[3], players[2], players[5]],
        [players[0], players[1], players[2], players[3], players[4], players[5]],
    ]
    return matchings


def swiss_bracket_n6(players: List[Any]) -> List[List[Any]]:
    '''
    Source : https://github.com/ValveSoftware/counter-strike_rules_and_regs/blob/main/major-supplemental-rulebook.md
    Look for the 'Priority' table in section 'Swiss Bracket'
    '''

    if len(players) != 6:
        msg = f'swiss_bracket_n6 function is implemented only for 6 participants, len(players) = {len(players)})'
        raise ValueError(msg)

    matchings = [
        [players[0], players[5], players[1], players[4], players[2], players[3]],
        [players[0], players[5], players[1], players[3], players[2], players[4]],
        [players[0], players[4], players[1], players[5], players[2], players[3]],
        [players[0], players[4], players[1], players[3], players[2], players[5]],
        [players[0], players[3], players[1], players[5], players[2], players[4]],

        [players[0], players[3], players[1], players[4], players[2], players[5]],
        [players[0], players[5], players[1], players[2], players[3], players[4]],
        [players[0], players[4], players[1], players[2], players[3], players[5]],
        [players[0], players[2], players[1], players[5], players[3], players[4]],
        [players[0], players[2], players[1], players[4], players[3], players[5]],

        [players[0], players[3], players[1], players[2], players[4], players[5]],
        [players[0], players[2], players[1], players[3], players[4], players[5]],
        [players[0], players[1], players[2], players[5], players[3], players[4]],
        [players[0], players[1], players[2], players[4], players[3], players[5]],
        [players[0], players[1], players[2], players[3], players[4], players[5]],
    ]
    return matchings



from typing import List, Dict

from rstt import Duel
from rstt.stypes import SPlayer
from . import Competition


class Snake(Competition):
    """The Snake Tournament

    This format is not a standard one. It is a model I detail and justify the interest in my master thesis and should be the subject of a paper
    in an hopefully near future.

    It does have some ressamblances with the `Ladder tournament <https://en.wikipedia.org/wiki/Ladder_tournament>`_.
    I am aware of two practical instances of uses:
        - The bonus round in the fencing challenge of the `modern pentathlon <https://en.wikipedia.org/wiki/Modern_pentathlon>`_ at the Paris Olympics 2024.
        - A cooking show with a `Hidden last chance <https://fr.wikipedia.org/wiki/Saison_14_de_Top_Chef>`_ secret tournament.

    Quick Overview:
        - For n participants it produces n-1 matches.
        - Every participants plays at least one match.
        - The final standing is an untied ranking.
        - Matches are all 'a priori balanced'.
    """

    def __init__(self, *args, **kwars):
        super().__init__(*args, **kwars)
        self.snake = []

    def _initialise(self) -> None:
        self.snake = [player for player in self.seeding]
        self.snake.reverse()

    def _update(self) -> None:
        self.snake.insert(0, self.played_matches[-1][0].winner())

    def _end_of_stage(self) -> bool:
        return len(self.snake) == 1

    def _standing(self) -> Dict[SPlayer, int]:
        standing = {games[0].loser(): len(self.participants())-r
                    for r, games in enumerate(self.played_matches)}
        standing[self.played_matches[-1][0].winner()] = 1
        return standing

    def generate_games(self) -> List[Duel]:
        return [Duel(self.snake.pop(0), self.snake.pop(0))]



import numpy as np


def flatten(list):
    # https://stackoverflow.com/questions/952914/how-do-i-make-a-flat-list-out-of-a-list-of-lists
    return [item for sublist in list for item in sublist]


def uniques(values: list):
    # inspired by:
    # https://stackoverflow.com/questions/61378055/how-to-find-values-repeated-more-than-n-number-of-times-using-only-numpy
    np_val = np.array(values)
    vals, counts = np.unique(np_val, return_counts=True)
    count = dict(zip(vals, counts))
    return set([key for key, value in count.items() if value == 1])


def multiples(values: list):
    return set([value for value in values if value not in uniques(values)])


def power_of_two(n):
    # https://stackoverflow.com/questions/57025836/how-to-check-if-a-given-number-is-a-power-of-two
    return (n != 0) and (n & (n-1) == 0)


def nmax(l: list, n: int):
    # https://stackoverflow.com/questions/50477976/location-of-n-max-values-in-a-python-list
    b = l[:]
    locations = []
    minimum = min(b)-1
    for i in range(n):
        maxIndex = b.index(max(b))
        locations.append(maxIndex)
        b[maxIndex] = minimum
    return [l[loc] for loc in locations]



"""Module for Observer

Observer is a protocol providing a handle_observations() method for ranking.
This method is responsible:

    - to proprocess observations - parameters passed to :func:`rstt.ranking.ranking.Raning.update` justifying ratings changes.
    - properly call the ranking.backend :func:`rstt.stypes.Infere.rate`
    - store new ratings in the ranking.datamodel :class:`rstt.stypes.RatingSystem`
    
.. warning::
    Currently RSTT SMatch support is limited to the Duel class.
    This is also true for 'Game based' Observer. Typecheckers may look for SMatch and not for Duel.
    This can cause unexpected errors in your simulations.

"""

from .obs import ObsTemplate, NoHandling
from .gameObserver import GameByGame, BatchGame
from .playerObserver import PlayerChecker



from typing import List
from typeguard import typechecked

from rstt import Duel
from rstt.stypes import SPlayer

import warnings


@typechecked
def playersToDuel(players: List[SPlayer]) -> List[Duel]:
    return [Duel(players[2*i], players[2*i+1]) for i in range(len(players)//2)]


@typechecked
def new_matchup(game: Duel, games: List[Duel], symetric: bool = True) -> bool:
    p1, p2 = game.player1(), game.player2()
    confrontations = [(duel.player1(), duel.player2()) for duel in games]
    if (p1, p2) in confrontations:
        return False
    if symetric and (p2, p1) in confrontations:
        return False
    return True


@typechecked
def find_valid_draw(draws: List[List[Duel]], games: List[Duel], symetric: bool = True) -> List[Duel]:
    # !!! symetric -> valid_draw(...) -> new_matchup(...)
    # avoid side-effect (usage of .pop)
    draws = [option for option in draws]

    # default return value
    first_option = draws[0]

    # find draw
    good_draw = None
    while not good_draw and draws:
        option = draws.pop(0)
        if valid_draw(option, games):
            good_draw = option

    # deal with mission failed
    if not good_draw:
        good_draw = first_option
        msg = f"No Valid matchups where found"
        warnings.warn(msg, RuntimeWarning)

    return good_draw


@typechecked
def valid_draw(draw: List[Duel], games: List[Duel]) -> bool:
    for game in draw:
        if not new_matchup(game, games):
            return False
    return True



from typing import Union, List, Dict, Tuple, Optional, Any, Iterable
from typing import Protocol, runtime_checkable
from typeguard import typechecked

from rstt import Duel, BetterWin
from rstt.ranking.ranking import Ranking
from rstt.stypes import SPlayer, Solver, Shuffler, Seeder, Generator, Evaluator
from . import Competition

from rstt.utils import utils as uu, matching as um, competition as uc


class DummyParam(Shuffler, Seeder, Evaluator):
    def rearange(status: List[int]) -> List[int]:
        return status

    def seed(players: List[SPlayer], *args,  **kwargs) -> List[SPlayer]:
        return players

    def generate(status: Union[List[int], List[SPlayer]]) -> Union[List[int], List[SPlayer]]:
        return um.ruban(status)

    def eval(options: List[List[SPlayer]], *args,  **kwargs) -> List[List[SPlayer]]:
        return options


class SwissBracket(Competition):
    @typechecked
    def __init__(self, name: str, seeding: Ranking, solver: Solver = BetterWin(),
                 matchings: Optional[Dict[Tuple[int, int], Shuffler]] = None,
                 seeders: Optional[Dict[Tuple[int, int], Seeder]] = None,
                 generators: Optional[Dict[Tuple[int, int], Generator]] = None,
                 evaluators: Optional[Dict[Tuple[int, int], Evaluator]] = None,
                 def_matching: Shuffler = DummyParam,
                 def_seeder: Seeder = DummyParam,
                 def_generator: Generator = DummyParam,
                 def_evaluator: Evaluator = DummyParam,
                 cashprize: Optional[Dict[int, float]] = None):
        """The Swiss Bracket

        The format is mention in the `esports variation <https://en.wikipedia.org/wiki/Swiss-system_tournament#Esports>`_ of the Swiss Round wikipedia page and was originaly introduce for Counter-Strike event in 2016.
        Hower both format differ drastically in many regards, one being their usage and output.
            * In Swiss Round there is a single undefeated participant, thus is great to crown a winner among many participants.
            * In Swiss Bracket, 16 teams are split into 8 qualified and 8 eliminated participants.

        The name choice come from the `Valve rule book <https://github.com/ValveSoftware/counter-strike_rules_and_regs/blob/main/major-supplemental-rulebook.md#swiss-bracket>`_ for Counter-Strike major competition, that specify a matching policy.

        Parameters
        ----------
        name : str
            A unique name to identify the Event.
        seeding : Ranking
            A ranking used for `seeding <https://en.wikipedia.org/wiki/Seeding_(sports)>`_ purposes.
        solver : Solver, optional
            A Solver to generate match outcomes, by default BetterWin()
        matchings : Optional[Dict[Tuple[int, int], Shuffler]], optional
            A collection of matching strategy for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the strategy is used.
        seeders : Optional[Dict[Tuple[int, int], Seeder]], optional
            A scollection of eeding strategy for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the strategy is used.
        generators : Optional[Dict[Tuple[int, int], Generator]], optional
            A collection of generator for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the generator is used.
        evaluators : Optional[Dict[Tuple[int, int], Evaluator]], optional
            A collection of evaluators for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the strategy is used.
        def_matching : Shuffler, optional
            A default matching strategy for rounds where the matchings param does not provide one, by default DummyParam
        def_seeder : Seeder, optional
            A default seeding strategy for rounds where the seeders param does not provide one, by default DummyParam
        def_generator : Generator, optional
            A default generator for rounds where the generator param does not provide one, by default DummyParam
        def_evaluator : Evaluator, optional
            A default evaluator for rounds where the evaluators param does not provide one, by default DummyParam
        cashprize : Optional[Dict[int, float]], optional
            A 'prizepool' rewarding player with 'money' for their success (placement in the final standing) during the Event, by default None

        .. note::
            The design of the class alows extreme customization on the matching at any point of the competition.
            If you do not provide a matching, seeders, genrators, evaluators and their associated default, the class uses a DummyParam for all.

            Refer to the :class:`rstt.stypes` documentation for the parameters type and impact on the tournament games.
        """
        super().__init__(name, seeding, solver, cashprize)

        # general settings of the system
        self.current_round = 0
        self.max_round = None
        self.max_wins = None
        self.max_loses = None

        # round parameters - map score with data: (int, int) -> data
        self.rounds = {}  # List[Player]
        self.matchings = matchings if matchings else {}  # Callable
        self.seeders = seeders if seeders else {}  # Callable
        self.generators = generators if generators else {}  # Callable
        self.evaluators = evaluators if evaluators else {}  # Callable

        self.def_matching = def_matching
        self.def_seeder = def_seeder
        self.def_generator = def_generator
        self.def_evaluator = def_evaluator

    # --- Override ---#
    def _initialise(self):
        # !!! Currently hardcoded setting for 16 participants
        msg = f"Currently SwissBracket is implemented only for 16 players, received {len(self.participants())}"
        assert len(self.participants()) == 16, msg

        self.__init_settings()
        self.__init_params()

    def _update(self) -> None:
        # ??? clear self.rounds or keep history
        # move winner and loser
        for game in self.played_matches[-1]:
            p1, p2 = game.player1(), game.player2()
            self.rounds[self.score(p1)].append(p1)
            self.rounds[self.score(p2)].append(p2)

        # move to next round
        self.current_round += 1

    def _end_of_stage(self) -> bool:
        return self.current_round == self.max_round

    def _standing(self) -> Dict[SPlayer, int]:
        standing = {}
        top8 = [(self.max_wins, i) for i in range(self.max_loses)]
        bottom8 = [(i, self.max_loses) for i in range(self.max_wins-1, -1, -1)]
        final_scores = top8 + bottom8
        top = 0
        for score in final_scores:
            players = self.rounds[score]
            top += len(players)
            for p in players:
                standing[p] = top
        return standing

    def generate_games(self):
        games = [uc.find_valid_draw(draws=self.draws(group=self.rounds[score]),
                                    games=self.games())
                 for score in self.round_scores()]
        return uu.flatten(games)

    # --- round mechanism --- #
    @typechecked
    def score(self, player: SPlayer) -> Tuple[int, int]:
        wins = len([game for game in self.games() if game.winner() == player])
        loses = len([game for game in self.games() if game.loser() == player])
        return (wins, loses)

    def round_scores(self) -> List[Tuple[int, int]]:
        # Round1 [(0, 0)]; Round2 [(1,0), (0,1)]; Round3 [(2,0), (1,1), (0,2)] ...
        return [(self.current_round-i, i) for i in range(self.current_round+1)
                if i < self.max_loses and self.current_round-i < self.max_wins]

    @typechecked
    def draws(self, group: List[SPlayer]) -> List[List[Duel]]:
        scores = [self.score(player) for player in group]
        # !!! debugging - SwissBracket will change to support different amount of players
        assert len(set(scores)) == 1
        score = scores[0]

        # sellect correct round parameters
        matching = self.matchings[score].rearange
        seeder = self.seeders[score].seed
        generator = self.generators[score].generate
        evaluator = self.evaluators[score].eval

        # how do seeds face each other by default
        paired_list = matching(list(range(len(group))))

        # who has which seeds - replace 'int' by 'Player'
        paired_player = seeder(players=[group[i] for i in paired_list],
                               initial_seeds=self.seeding,
                               results=self.results())

        # what are all the options we want to check
        options = generator(paired_player)

        # in which order should we consider them
        options = evaluator(options,
                            initial_seeds=self.seeding,
                            results=self.results())

        return [uc.playersToDuel(option) for option in options]

    def results(self):
        # fit data to expected format for tiebreaker function
        res = {player: {'opponent': [], 'score': [], }
               for player in self.seeding}

        for game in self.games():
            # update opponent list
            res[game.player1()]['opponent'].append(game.player2())
            res[game.player2()]['opponent'].append(game.player1())

            # update score list !!! Lacks generalisation
            res[game.player1()]['score'].append(
                'win' if game.winner() == game.player1() else 'lose')
            res[game.player2()]['score'].append(
                'win' if game.winner() == game.player2() else 'lose')
        return res

    # --- under the hood mechanism --- #
    def __possible_scores(self):
        return [(i, j) for j in range(self.max_loses+1) for i in range(self.max_wins+1) if i+j <= self.max_round]

    def __init_settings(self):
        self.max_round = 5
        self.max_wins = 3
        self.max_loses = 3

    def __init_params(self):
        scores = self.__possible_scores()

        self.rounds[(0, 0)] = [player for player in self.seeding]
        self.rounds = {score: self.rounds[score] if score in self.rounds.keys()
                       else [] for score in scores}

        self.matchings = {score: self.matchings[score] if score in self.matchings.keys()
                          else self.def_matching for score in scores}

        self.seeders = {score: self.seeders[score] if score in self.seeders.keys()
                        else self.def_seeder for score in scores}

        self.generators = {score: self.generators[score] if score in self.generators.keys()
                           else self.def_generator for score in scores}

        self.evaluators = {score: self.evaluators[score] if score in self.evaluators.keys()
                           else self.def_evaluator for score in scores}



from typing import List, Union, Optional
from typeguard import typechecked

from rstt.player.basicplayer import BasicPlayer
from rstt.game import Match
from rstt.stypes import SMatch, Achievement


class Player(BasicPlayer):
    @typechecked
    def __init__(self, name: Optional[str] = None, level: Optional[float] = None) -> None:
        """Player with match history.

        Player extends :class:`rstt.player.basicPlayer.BasicPlayer` with the possibility to track games it played in and
        by collecting results achieved in :class:`rstt.stypes.Achievement`. 

        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        level : float, optional
            The level/skill/strenght of the player. By default None, in this case a level is randomly generated.
            using a random distribution :class:`rstt.config.PLAYER_DIST
            with default parameters :class:`rstt.config.PLAYER_DIST_ARGS`
        """
        super().__init__(name=name, level=level)
        self.__achievements: list[Achievement] = []
        self.__games: list[SMatch] = []

    # --- getter --- #
    def achievements(self) -> List[Achievement]:
        """Getter method for achievement

        :class:`rstt.stypes.Achievement` represent tournament result of the player. 

        Returns
        -------
        List[Achievement]
            All past success of the player in chronological order, from the oldest to the most recent.
        """
        return self.__achievements

    def earnings(self) -> float:
        """Getter method for the earnings

        Sugar method that returns the sum of earnings specified by the player achievements.

        Returns
        -------
        float
            All the money earned in competitif event.
        """
        return sum([achievement.prize for achievement in self.__achievements])

    def games(self) -> List[Match]:
        """Getter method for match the player participated in

        Returns
        -------
        List[Match]
            All the matches the player played in chronolgical order, from oldest to the most recent.
        """
        return self.__games

    # --- setter --- #
    @typechecked
    def collect(self, achievement: Union[Achievement, List[Achievement]]):
        """Adds achivement(s) to the player history.

        Parameters
        ----------
        achievement : Union[Achievement, List[Achievement]]
            Achievement(s) passed must have a different event_name attribute that the one already stored.

        Raises
        ------
        ValueError
            Raised when attempting to collect an event with an event_name already present in the player history. 
        """
        if isinstance(achievement, Achievement):
            achievements = [achievement]
        else:
            achievements = achievement

        previous_event = [
            past_event.event_name for past_event in self.__achievements]
        for achievement in achievements:
            if achievement.event_name not in previous_event:
                self.__achievements.append(achievement)
            else:
                msg = f"Can not collect {achievement}. {self} already participated in an event called {achievement.event_name}"
                raise ValueError(msg)

    @typechecked
    def add_game(self, match: Match) -> None:
        """Adds match to the player history

        Parameters
        ----------
        match : Match
            A match to track.

        Raises
        ------
        ValueError
            The match needs to be a game in which the player partipiated in and not already tracked. Either condition violated will raise an Error.
        """
        if match in self.__games:
            msg = f"{match} already present in game history of player {self}"
            raise ValueError(msg)
        if self not in match.players():  # !!! Untested Feature
            msg = f"Can not collect a match {match} that player {self} has not been part of."
            raise ValueError
        self.__add_game(match)

    @typechecked
    def reset(self, games: bool = True, achievement: bool = True) -> None:
        """Clean the player history

        Removes all matchs and achievements from the players history. 

        Parameters
        ----------
        games : bool, optional
            Wether matchs should be removed, by default True. 
        achievement : bool, optional
            Wether achievements should be removed, by default True.
        """
        if achievement:
            self.__achievements = []
        if games:
            self.__games = []

    # --- internal mechanism --- #
    def __add_game(self, match: Match) -> None:
        self.__games.append(match)
